# Import preprocessing functions
from preprocessing import (
    tfidf_prepro, encoder_prepro, encoder_arabe_prepro, 
    tfidf_latin_prepro, tfidf_arabe_prepro, encoder_latin_prepro
)

# Import evaluation and training utilities
from evaluate import tfdidf_eval
from train import train_network
from create_submission import create_output

# Import models
from models import clf, TransformerClassifier, xgb_model, SimpleClassifier

# Cross-validation and language analysis
from cross_val import XGB_crossval

# Standard libraries and machine learning tools
import pandas as pd
import torch

# ----------------------------------------
# TF-IDF Baseline Training and Evaluation
# ----------------------------------------
def main_tfidf():
    X_train, y_train, X_val, y_val, X_test, y_test, vectorizer = tfidf_prepro()
    clf.fit(X_train, y_train)

    print("Hello from anlp-kaggle!")
    tfdidf_eval(X_val, y_val, clf)
    create_output(vectorizer, clf)


# ----------------------------------------
# Cross-validation with XGBoost
# ----------------------------------------
def main_cv():
    X_train, y_train, X_val, y_val, X_test, y_test, vectorizer = tfidf_prepro()
    print(len(set(y_train)), len(set(y_val)))  # Log unique class counts
    XGB_crossval(X_train, y_train, X_val, y_val, xgb_model)


# ----------------------------------------
# Transformer-based Encoder Training
# ----------------------------------------
def main_transfo_encoder():
    X_train, y_train, X_val, y_val, X_test, y_test = encoder_prepro()

    # Convert data to tensors
    X_train = torch.tensor(X_train, dtype=torch.long)
    y_train = torch.tensor(y_train, dtype=torch.long)
    X_val = torch.tensor(X_val, dtype=torch.long)
    y_val = torch.tensor(y_val, dtype=torch.long)
    X_test = torch.tensor(X_test, dtype=torch.long)
    y_test = torch.tensor(y_test, dtype=torch.long)

    print(X_train.size(), y_train.size())

    # Define transformer model parameters
    vocab_size = 50000
    embedding_dim = 128
    num_heads = 8
    hidden_dim = 256
    num_layers = 1
    num_classes = 382

    model = TransformerClassifier(vocab_size, embedding_dim, num_heads, hidden_dim, num_layers, num_classes)
    train_network(model, X_train, y_train, X_val, y_val)


# ----------------------------------------
# Arabic Encoder Training
# ----------------------------------------
def main_arabe_encoder():
    X_train, y_train, X_val, y_val, X_test, y_test = encoder_arabe_prepro()
    clf.fit(X_train, y_train)

    print("Hello from anlp-kaggle!")
    tfdidf_eval(X_val, y_val, clf)
    # create_output(vectorizer, clf)  # Optional


# ----------------------------------------
# Latin Encoder Training
# ----------------------------------------
def main_latin_encoder():
    X_train, y_train, X_val, y_val, X_test, y_test = encoder_latin_prepro()
    clf.fit(X_train, y_train)

    print("Hello from anlp-kaggle!")
    tfdidf_eval(X_val, y_val, clf)
    # create_output(vectorizer, clf)  # Optional


# ----------------------------------------
# Latin Classifier with Neural Network
# ----------------------------------------
def main_latin_classifier_nn():
    X_train, y_train, X_val, y_val, X_test, y_test = encoder_latin_prepro()

    num_classes = len(set(y_train.tolist()))

    clf = SimpleClassifier(input_dim=X_train.shape[1], hidden_dim=512, num_classes=num_classes)
    train_network(clf, X_train, y_train, X_val, y_val)


# ----------------------------------------
# Arabic TF-IDF Training and Evaluation
# ----------------------------------------
def main_arabe_tfidf():
    X_train, y_train, X_val, y_val, X_test, y_test, vectorizer = tfidf_arabe_prepro()
    clf.fit(X_train, y_train)

    tfdidf_eval(X_val, y_val, clf)
    # create_output(vectorizer, clf)  # Optional


# ----------------------------------------
# Latin TF-IDF Training and Evaluation
# ----------------------------------------
def main_latin_tfidf():
    X_train, y_train, X_val, y_val, X_test, y_test, vectorizer = tfidf_latin_prepro()

    num_classes = len(set(y_train))

    # Convert sparse matrices to dense tensors
    X_train = torch.tensor(X_train.toarray(), dtype=torch.float)
    X_val = torch.tensor(X_val.toarray(), dtype=torch.float)
    X_test = torch.tensor(X_test.toarray(), dtype=torch.float)

    y_train = torch.tensor(y_train, dtype=torch.long)
    y_val = torch.tensor(y_val, dtype=torch.long)
    y_test = torch.tensor(y_test, dtype=torch.long)

    clf = SimpleClassifier(input_dim=X_train.shape[1], hidden_dim=128, num_classes=num_classes)
    train_network(clf, X_train, y_train, X_val, y_val)
    # create_output(vectorizer, clf)  # Optional


# ----------------------------------------
# Arabic Classifier with Neural Network
# ----------------------------------------
def main_arabe_classifier_nn():
    X_train, y_train, X_val, y_val, X_test, y_test, vectorizer = tfidf_arabe_prepro()

    # Convert sparse to dense
    X_train = torch.tensor(X_train.toarray(), dtype=torch.float)
    X_val = torch.tensor(X_val.toarray(), dtype=torch.float)
    X_test = torch.tensor(X_test.toarray(), dtype=torch.float)

    y_train = torch.tensor(y_train, dtype=torch.long)
    y_val = torch.tensor(y_val, dtype=torch.long)
    y_test = torch.tensor(y_test, dtype=torch.long)

    num_classes = len(set(y_train.tolist()))
    clf = SimpleClassifier(input_dim=X_train.shape[1], hidden_dim=1024, num_classes=num_classes)

    train_network(clf, X_train, y_train, X_val, y_val)



# Run desired main function
if __name__ == "__main__":


    main_latin_tfidf()