import pandas as pd
import numpy as np
import re
import torch

from pathlib import Path
from tokenizers import Tokenizer
from tokenizers.models import BPE
from tokenizers.trainers import BpeTrainer
from tokenizers.pre_tokenizers import Whitespace
from transformers import AutoTokenizer, AutoModel


from sklearn.model_selection import train_test_split
from sklearn.feature_extraction.text import TfidfVectorizer
from tqdm import tqdm



DATA_PATH = Path("data") 


def tfidf_prepro():
    # Load the dataset
    train_raw = pd.read_csv(DATA_PATH / "train_submission_clean.csv")
    
    # Drop the "Usage" column if it exists — it's not needed for training
    train_raw.drop(columns=["Usage"], inplace=True)
    
    # Convert the 'Label' column to a categorical type
    train_raw["Label"] = train_raw["Label"].astype("category")
    
    # Remove any rows with missing values (NaN)
    train_raw = train_raw.dropna()
    
    # Get class counts to filter rare labels (less than 10 examples)
    class_counts = train_raw['Label'].value_counts()
    valid_classes = class_counts[class_counts >= 10].index
    train_raw = train_raw[train_raw['Label'].isin(valid_classes)]
    
    # Compute text length and filter out texts longer than 200 characters
    train_raw['Text_length'] = train_raw['Text'].apply(len)
    train_raw = train_raw[train_raw['Text_length'] <= 200]
    
    # Extract the texts and labels
    texts = train_raw["Text"]
    
    # Map labels to integer indices (for training compatibility)
    label_mapping = {label: idx for idx, label in enumerate(train_raw['Label'].unique())}
    train_raw['Label'] = train_raw['Label'].map(label_mapping)
    labels = train_raw["Label"]
    
    # Split the data into training (80%) and validation (20%) sets
    X_train_texts, X_val_texts, y_train, y_val = train_test_split(
        texts, 
        labels, 
        test_size=0.2,        # 20% for validation
        random_state=42,      # Set random state for reproducibility
        stratify=labels        # Ensure class distribution is maintained
    )
    
    # Further split validation into validation (70%) and test (30%) sets
    X_val_texts, X_test_texts, y_val, y_test = train_test_split(
        X_val_texts, 
        y_val, 
        test_size=0.3,        # 30% of validation set becomes test set (final ~6% of total data)
        random_state=42,      # Same random state for consistency
        stratify=y_val         # Ensure class distribution is maintained
    )
    
    # Create a character-level TF-IDF vectorizer
    vectorizer = TfidfVectorizer(
        analyzer='char',        # Analyze at the character level (better for noisy short text)
        ngram_range=(1,5),      # Use character n-grams from size 1 to 5
        max_features=50000,     # Limit to top 50,000 features to save memory
        lowercase=True          # Convert all text to lowercase
    )

    # Fit the vectorizer on training texts and transform train/val/test sets
    X_train = vectorizer.fit_transform(X_train_texts)
    X_val = vectorizer.transform(X_val_texts)
    X_test = vectorizer.transform(X_test_texts)

    # Return all the processed data and the vectorizer itself
    return X_train, y_train, X_val, y_val, X_test, y_test, vectorizer


def encoder_prepro():
    # Load the dataset from CSV file
    train_raw = pd.read_csv(DATA_PATH / "train_submission_clean.csv")
    
    # Drop the "Usage" column since it's not needed
    train_raw.drop(columns=["Usage"], inplace=True)
    
    # Convert 'Label' column to categorical type for easier processing
    train_raw["Label"] = train_raw["Label"].astype("category")
    
    # Remove rows with missing data
    train_raw = train_raw.dropna()

    # Get count of each label
    class_counts = train_raw['Label'].value_counts()

    # Keep only labels with at least 10 examples
    valid_classes = class_counts[class_counts >= 10].index
    train_raw = train_raw[train_raw['Label'].isin(valid_classes)]

    # Compute text length for each example
    train_raw['Text_length'] = train_raw['Text'].apply(len)

    # Filter out texts longer than 200 characters
    train_raw = train_raw[train_raw['Text_length'] <= 200]

    # Extract texts (input) and labels (output)
    texts = train_raw["Text"]

    # Map class labels to integer indices
    label_mapping = {label: idx for idx, label in enumerate(train_raw['Label'].unique())}
    train_raw['Label'] = train_raw['Label'].map(label_mapping)
    labels = train_raw["Label"]

    # Split into train (80%) and validation (20%)
    X_train_texts, X_val_texts, y_train, y_val = train_test_split(
        texts,
        labels,
        test_size=0.2,        # 20% for validation
        random_state=42,      # Set seed for reproducibility
        stratify=labels        # Maintain class distribution across splits
    )

    # Further split validation into validation (70%) and test (30%)
    X_val_texts, X_test_texts, y_val, y_test = train_test_split(
        X_val_texts,
        y_val,
        test_size=0.3,        # 30% of val becomes test (final ~6% of data)
        random_state=42,
        stratify=y_val        # Maintain class distribution across splits
    )

    # Load the pre-trained tokenizer (assumes BPE tokenizer trained separately)
    tokenizer = Tokenizer.from_file("bpe_encoder_50000.json")

    # Apply the tokenizer to convert texts into sequences of token IDs
    X_train_tokenized = [tokenizer.encode(text).ids for text in X_train_texts]
    X_val_tokenized = [tokenizer.encode(text).ids for text in X_val_texts]
    X_test_tokenized = [tokenizer.encode(text).ids for text in X_test_texts]

    # Determine max sequence length across all splits (train, val, test)
    max_len = max(
        max(len(seq) for seq in X_train_tokenized),
        max(len(seq) for seq in X_val_tokenized),
        max(len(seq) for seq in X_test_tokenized)
    )

    # Function to pad sequences to the max length
    def pad_sequences(sequences, max_len, padding_value=-1):
        padded = []
        for seq in sequences:
            if len(seq) < max_len:
                # Pad the sequence with -1 until max length is reached
                seq = seq + [-1] * (max_len - len(seq))
            padded.append(seq)
        return torch.tensor(padded, dtype=torch.long)

    # Pad all tokenized sequences to uniform length (required for batching)
    X_train = pad_sequences(X_train_tokenized, max_len)
    X_val = pad_sequences(X_val_tokenized, max_len)
    X_test = pad_sequences(X_test_tokenized, max_len)

    # Convert labels to PyTorch tensors (already class indices)
    y_train = y_train.values.astype(np.int64)
    y_train = torch.tensor(y_train, dtype=torch.long)

    y_val = y_val.values.astype(np.int64)
    y_val = torch.tensor(y_val, dtype=torch.long)

    y_test = y_test.values.astype(np.int64)
    y_test = torch.tensor(y_test, dtype=torch.long)

    # Optional sanity check: Print shapes of all datasets
    print(f"X_train shape: {X_train.shape}, y_train shape: {y_train.shape}")
    print(f"X_val shape: {X_val.shape}, y_val shape: {y_val.shape}")
    print(f"X_test shape: {X_test.shape}, y_test shape: {y_test.shape}")

    # Return processed data and labels
    return X_train, y_train, X_val, y_val, X_test, y_test

def remove_links(text):
    # Regex pattern to match most common URL formats (http, https, www)
    url_pattern = r'http[s]?://\S+|www\.\S+'
    
    # Remove URLs from text
    cleaned_text = re.sub(url_pattern, '', text)
    
    return cleaned_text.strip()

to_remove = ["#", "*", "+", "-", "/", "<", "=", ">", "@", "[", "\\", "]", "^", "_", "`", "{", "|", "}", "~", "", "", "", "", "¤", "¥", "§", "©", "ª", "«", "¬", "®", "¯", "°", "±", "²", "³", "´", "¶", "¹", "º", "»", "½", "¾", "ʰ", "ʲ", "ˀ", "˻", "˼"]


def remove_lines_with_chars_and_links(csv_file, chars_to_remove, output_file = 'train_submission_clean.csv', column_name = 'Text'):
    """
    Reads a CSV file, filters out lines in the specified column that contain any of the characters in chars_to_remove,
    and writes the remaining lines to a new CSV file.
    
    Args:
    csv_file (str or Path): Path to the input CSV file.
    column_name (str): Name of the column to check for characters.
    chars_to_remove (str): String containing characters that, if present in a line, will cause the line to be removed.
    output_file (str or Path): Path to the output CSV file where the filtered data will be saved.
    """
    # Load the data
    df = pd.read_csv(csv_file)
    df['Text'].apply(remove_links)
    # Remove specified characters from the text in the column
    df[column_name] = df[column_name].astype(str).apply(lambda x: ''.join(char for char in x if char not in chars_to_remove))
    
    # Save the filtered data to a new CSV file
    df.to_csv('data/'+output_file, index=False)
    print(f"Original number of lines: {len(df)}")
    print(f"Number of lines after removal: {len(df)}")


def bpe_encoder():

    # Initialiser un tokenizer avec un modèle BPE
    tokenizer = Tokenizer(BPE())
    tokenizer.pre_tokenizer = Whitespace()

    # Entraîner le tokenizer sur un corpus
    trainer = BpeTrainer(vocab_size=50000, show_progress=True)
    tokenizer.train(["data/corpus.txt"], trainer)

    # Enregistrer le tokenizer
    tokenizer.save("bpe_encoder_50000.json")
     

def extract_texts_to_txt(csv_file, column_name, output_file):
    # Lire le fichier CSV
    df = pd.read_csv(csv_file / 'train_submission_clean.csv')

    # Vérifier si la colonne existe
    if column_name not in df.columns:
        raise ValueError(f"La colonne '{column_name}' n'existe pas dans le fichier CSV.")

    # Extraire les textes de la colonne spécifiée
    texts = df[column_name].dropna().tolist()

    # Écrire les textes dans un fichier texte
    with open(output_file, 'w', encoding='utf-8') as f:
        for text in texts:
            f.write(text + '\n')

'''# Exemple d'utilisation
csv_file = DATA_PATH # Remplacez par le nom de votre fichier CSV
column_name = 'Text'  # Remplacez par le nom de la colonne contenant les textes
output_file = 'data/corpus.txt'  # Nom du fichier texte de sortie

extract_texts_to_txt(csv_file, column_name, output_file)'''


def is_arabic(text, threshold=0.5):
    arabic_characters = set(
        "ءآأؤإئابتثجحخدذرزسشصضطظعغفقكلمنهوىيًٌٍَُِّْ"
    )
    arabic_count = sum(1 for char in text if char in arabic_characters)
    ratio = arabic_count / len(text) if len(text) > 0 else 0
    return ratio >= threshold


def is_latin(text, threshold=0.5):
    latin_characters = set(
        "abcdefghijklmnopqrstuvwxyz"
    )
    latin_count = sum(1 for char in text if char in latin_characters)
    ratio = latin_count / len(text) if len(text) > 0 else 0
    return ratio >= threshold


def encode_arabic_texts(text_list):
    # Handle empty input list: return an empty tensor with shape (0, 768)
    if len(text_list) == 0:
        return torch.empty((0, 768))

    # Set the device: use Apple MPS if available (for Macs with Metal support), otherwise use CPU
    device = torch.device("mps") if torch.backends.mps.is_available() else torch.device("cpu")

    # Load AraBERT tokenizer and model (only loaded once for efficiency)
    tokenizer = AutoTokenizer.from_pretrained("aubmindlab/bert-base-arabertv02")
    model = AutoModel.from_pretrained("aubmindlab/bert-base-arabertv02").to(device)

    # Tokenize input text list: 
    # - Add padding to ensure uniform input size
    # - Truncate if text is too long
    # - Set max length to 512 (BERT max token length)
    # - Return tensors for PyTorch
    inputs = tokenizer(
        text_list,
        padding=True,
        truncation=True,
        max_length=512,
        return_tensors="pt"
    ).to(device)

    # Perform inference without calculating gradients (faster and uses less memory)
    with torch.no_grad():
        outputs = model(**inputs)

    # Compute sentence embeddings by averaging the hidden states across all tokens
    # Move back to CPU for further processing if needed
    embeddings = outputs.last_hidden_state.mean(dim=1).cpu()

    # Return the final embeddings (one vector per input text)
    return embeddings


def encode_latin_texts(text_list, batch_size=8):
    # Handle case where text list is empty — return empty tensor with shape (0, 256)
    if len(text_list) == 0:
        return torch.empty((0, 256))

    # Set device to CPU (MPS is checked but still falls back to CPU here)
    device = torch.device("cpu") if torch.backends.mps.is_available() else torch.device("cpu")

    # Load tokenizer and model from the sentence-transformers library
    # This model is suitable for short English/Latin texts (MiniLM is small and fast)
    tokenizer = AutoTokenizer.from_pretrained("sentence-transformers/paraphrase-MiniLM-L3-v2")
    model = AutoModel.from_pretrained("sentence-transformers/paraphrase-MiniLM-L3-v2").to(device)

    all_embeddings = []

    # Process texts in batches — useful for larger datasets to avoid memory issues
    # tqdm adds a progress bar to show encoding progress
    for i in tqdm(range(0, len(text_list), batch_size), desc="Encoding Latin Texts"):
        batch = text_list[i:i+batch_size]

        # Tokenize the batch:
        # - Add padding so all sequences in the batch have the same length
        # - Truncate if sequences exceed 256 tokens
        # - Return tensors for PyTorch
        inputs = tokenizer(
            batch,
            padding=True,
            truncation=True,
            max_length=256,  # Adjust if texts are very short
            return_tensors="pt"
        ).to(device)

        # Get embeddings from the model (no gradient needed for inference)
        with torch.no_grad():
            outputs = model(**inputs)

        # Get sentence-level embeddings by averaging the token embeddings
        embeddings = outputs.last_hidden_state.mean(dim=1).cpu()  # Move to CPU for compatibility

        # Collect embeddings for all batches
        all_embeddings.append(embeddings)

    # Concatenate all batch embeddings into a single tensor (final result)
    return torch.cat(all_embeddings, dim=0)

def encoder_latin_prepro():
    # Load the cleaned training data
    train_raw = pd.read_csv(DATA_PATH / "train_submission_clean.csv")
    
    # Drop the 'Usage' column — not needed for this preprocessing
    train_raw.drop(columns=["Usage"], inplace=True)
    
    # Convert labels to categorical type for easier handling
    train_raw["Label"] = train_raw["Label"].astype("category")
    
    # Drop any rows with missing values
    train_raw = train_raw.dropna()

    # Enable tqdm progress bar for pandas operations
    tqdm.pandas()
    
    # Print the MPS (Metal Performance Shaders) availability for Apple devices
    print(torch.backends.mps.is_available())
    print(torch.backends.mps.is_built())

    # Detect if each text is Latin (based on character detection using `is_latin()`)
    train_raw['is_latin'] = train_raw['Text'].progress_apply(is_latin)

    # Keep only Latin texts (filter non-Latin rows)
    train_raw = train_raw[train_raw['is_latin']]
    print("nombre de lignes latines :", len(train_raw))

    # Remove classes with fewer than 10 examples (rare classes are excluded)
    class_counts = train_raw['Label'].value_counts()
    valid_classes = class_counts[class_counts >= 10].index
    train_raw = train_raw[train_raw['Label'].isin(valid_classes)]

    # Compute text length and filter out very long texts (more than 200 characters)
    train_raw['Text_length'] = train_raw['Text'].apply(len)
    train_raw = train_raw[train_raw['Text_length'] <= 200]

    # Extract texts and labels
    texts = train_raw["Text"].tolist()

    # Map labels to integer indices for classification
    label_mapping = {label: idx for idx, label in enumerate(train_raw['Label'].unique())}
    train_raw['Label'] = train_raw['Label'].map(label_mapping)
    labels = train_raw["Label"].tolist()

    # Split data into training and validation sets (80% train, 20% validation)
    X_train_texts, X_val_texts, y_train, y_val = train_test_split(
        texts, labels,
        test_size=0.2,
        random_state=42,
        stratify=labels  # Ensure class distribution is preserved
    )

    # Split validation set into validation and test sets (70% val, 30% test)
    X_val_texts, X_test_texts, y_val, y_test = train_test_split(
        X_val_texts, y_val,
        test_size=0.3,
        random_state=42,
        stratify=y_val  # Ensure class distribution is preserved
    )

    # Encode the training, validation, and test texts using the Latin encoder
    X_train = encode_latin_texts(X_train_texts)
    X_val = encode_latin_texts(X_val_texts)
    X_test = encode_latin_texts(X_test_texts)

    # Convert labels to PyTorch tensors (needed for training)
    y_train = torch.tensor(y_train, dtype=torch.long)
    y_val = torch.tensor(y_val, dtype=torch.long)
    y_test = torch.tensor(y_test, dtype=torch.long)

    # Print shape of training set to confirm successful encoding
    print('print(X_train.shape)', X_train.shape)

    # Return encoded texts and labels for train, val, and test sets
    return X_train, y_train, X_val, y_val, X_test, y_test


def encoder_arabe_prepro():
    # Load the cleaned training data
    train_raw = pd.read_csv(DATA_PATH / "train_submission_clean.csv")
    
    # Drop the 'Usage' column — not needed for this preprocessing
    train_raw.drop(columns=["Usage"], inplace=True)

    # Convert the 'Label' column into categorical type (helps with mapping later)
    train_raw["Label"] = train_raw["Label"].astype("category")

    # Remove rows with any missing values
    train_raw = train_raw.dropna()

    # Count occurrences of each label
    class_counts = train_raw['Label'].value_counts()

    # Keep only labels (classes) that appear at least 10 times
    valid_classes = class_counts[class_counts >= 10].index
    train_raw = train_raw[train_raw['Label'].isin(valid_classes)]

    # Compute text lengths
    train_raw['Text_length'] = train_raw['Text'].apply(len)

    # Filter to keep only texts that are 200 characters or fewer
    train_raw = train_raw[train_raw['Text_length'] <= 200]

    # Enable progress bar for pandas operations using tqdm
    tqdm.pandas()

    # Print MPS availability (for Mac hardware acceleration)
    print(torch.backends.mps.is_available())
    print(torch.backends.mps.is_built())

    # Detect if each text is Arabic (based on Arabic character detection via `is_arabic()`)
    train_raw['is_arabic'] = train_raw['Text'].progress_apply(is_arabic)

    # Print number of Arabic rows found
    print("nombre de lignes arabes :", len(train_raw))

    # Keep only Arabic texts (filter out non-Arabic rows)
    train_raw = train_raw[train_raw['is_arabic']]

    # Extract the actual text content into a list
    texts = train_raw["Text"].tolist()

    # Create a mapping from label names to numeric indices
    label_mapping = {label: idx for idx, label in enumerate(train_raw['Label'].unique())}

    # Replace original labels with their corresponding numeric indices
    train_raw['Label'] = train_raw['Label'].map(label_mapping)

    # Extract labels into a list (now as numeric indices)
    labels = train_raw["Label"].tolist()

    # Split data into training and validation sets (80% train, 20% validation)
    X_train_texts, X_val_texts, y_train, y_val = train_test_split(
        texts, labels,
        test_size=0.2,
        random_state=42,
        stratify=labels  # Ensures balanced class distribution in both sets
    )

    # Split validation set into validation and test sets (70% val, 30% test)
    X_val_texts, X_test_texts, y_val, y_test = train_test_split(
        X_val_texts, y_val,
        test_size=0.3,
        random_state=42,
        stratify=y_val  # Ensures balanced class distribution in both sets
    )

    # Encode the Arabic texts into numerical embeddings using AraBERT
    X_train = encode_arabic_texts(X_train_texts)
    X_val = encode_arabic_texts(X_val_texts)
    X_test = encode_arabic_texts(X_test_texts)

    # Convert labels to PyTorch tensors (needed for training the model later)
    y_train = torch.tensor(y_train, dtype=torch.long)
    y_val = torch.tensor(y_val, dtype=torch.long)
    y_test = torch.tensor(y_test, dtype=torch.long)

    # Print the shape of the encoded training set (for debugging/confirmation)
    print('print(X_train.shape)', X_train.shape)

    # Return the processed and encoded data (for train, val, and test sets)
    return X_train, y_train, X_val, y_val, X_test, y_test


def tfidf_arabe_prepro():
    # Load the cleaned training dataset
    train_raw = pd.read_csv(DATA_PATH / "train_submission_clean.csv")
    
    # Drop the 'Usage' column since it's not needed for training
    train_raw.drop(columns=["Usage"], inplace=True)

    # Convert the 'Label' column into a categorical data type
    train_raw["Label"] = train_raw["Label"].astype("category")

    # Remove rows with missing values (if any)
    train_raw = train_raw.dropna()

    # Enable progress bars for pandas operations (using tqdm)
    tqdm.pandas()

    # Apply the Arabic detection function to each text in the dataset
    train_raw['is_arabic'] = train_raw['Text'].progress_apply(is_arabic)

    # Keep only rows where the text was detected as Arabic
    train_raw = train_raw[train_raw['is_arabic']]

    # Print the filtered dataset (for debugging purposes)
    print(train_raw)

    # Count how many times each label appears
    class_counts = train_raw['Label'].value_counts()

    # Only keep classes that appear at least 10 times
    valid_classes = class_counts[class_counts >= 10].index
    train_raw = train_raw[train_raw['Label'].isin(valid_classes)]

    # Compute the length of each text (number of characters)
    train_raw['Text_length'] = train_raw['Text'].apply(len)

    # Keep only texts that are 200 characters or fewer
    train_raw = train_raw[train_raw['Text_length'] <= 200]

    # Print the final number of Arabic texts kept
    print(len(train_raw))

    # Print MPS (Apple GPU acceleration) availability for debugging
    print(torch.backends.mps.is_available())
    print(torch.backends.mps.is_built())

    # Extract the texts into a list (needed for vectorization)
    texts = train_raw["Text"].tolist()

    # Create a mapping from label names to numeric indices
    label_mapping = {label: idx for idx, label in enumerate(train_raw['Label'].unique())}

    # Apply label mapping to the 'Label' column
    train_raw['Label'] = train_raw['Label'].map(label_mapping)

    # Extract the mapped numeric labels into a list
    labels = train_raw["Label"].tolist()

    # Split the dataset into train and validation (80% train, 20% validation)
    X_train_texts, X_val_texts, y_train, y_val = train_test_split(
        texts, labels,
        test_size=0.2,
        random_state=42,
        stratify=labels  # Ensures class balance in both splits
    )

    # Further split the validation set into validation and test (70% val, 30% test)
    X_val_texts, X_test_texts, y_val, y_test = train_test_split(
        X_val_texts, y_val,
        test_size=0.3,
        random_state=42,
        stratify=y_val  # Ensures class balance in both splits
    )

    # Initialize TF-IDF vectorizer with character-level analysis (character n-grams)
    vectorizer = TfidfVectorizer(
        analyzer='char',       # analyze text at character level
        ngram_range=(1,5),     # use n-grams from size 1 to 5
        max_features=500,      # limit to the top 500 features
        lowercase=True         # convert all text to lowercase
    )

    # Fit the vectorizer on the training texts and transform them into numerical vectors
    X_train = vectorizer.fit_transform(X_train_texts)

    # Transform the validation and test sets using the already fitted vectorizer
    X_val = vectorizer.transform(X_val_texts)
    X_test = vectorizer.transform(X_test_texts)

    # Return the TF-IDF matrices, labels, and the vectorizer itself
    return X_train, y_train, X_val, y_val, X_test, y_test, vectorizer


def tfidf_latin_prepro():
    # Load the cleaned training dataset
    train_raw = pd.read_csv(DATA_PATH / "train_submission_clean.csv")

    # Drop the 'Usage' column as it's irrelevant for this preprocessing
    train_raw.drop(columns=["Usage"], inplace=True)

    # Convert the 'Label' column into a categorical type
    train_raw["Label"] = train_raw["Label"].astype("category")

    # Drop any rows with missing values
    train_raw = train_raw.dropna()

    # Enable tqdm progress bar for pandas apply functions
    tqdm.pandas()

    # Apply the 'is_latin' function to detect Latin texts
    train_raw['is_latin'] = train_raw['Text'].progress_apply(is_latin)

    # Filter to keep only rows with Latin texts
    train_raw = train_raw[train_raw['is_latin']]

    # Display the filtered dataset (for debugging purposes)
    print(train_raw)

    # Count occurrences of each label
    class_counts = train_raw['Label'].value_counts()

    # Only keep labels that appear at least 10 times
    valid_classes = class_counts[class_counts >= 10].index
    train_raw = train_raw[train_raw['Label'].isin(valid_classes)]

    # Compute text lengths (number of characters)
    train_raw['Text_length'] = train_raw['Text'].apply(len)

    # Filter out texts longer than 200 characters
    train_raw = train_raw[train_raw['Text_length'] <= 200]

    # Print final count of Latin texts kept
    print(len(train_raw))

    # Print MPS (Apple GPU acceleration) availability for debugging
    print(torch.backends.mps.is_available())
    print(torch.backends.mps.is_built())

    # Extract the actual texts into a list
    texts = train_raw["Text"].tolist()

    # Create a label-to-index mapping
    label_mapping = {label: idx for idx, label in enumerate(train_raw['Label'].unique())}

    # Convert label names to numeric indices
    train_raw['Label'] = train_raw['Label'].map(label_mapping)

    # Extract labels into a list
    labels = train_raw["Label"].tolist()

    # Split texts and labels into training (80%) and validation (20%) sets
    X_train_texts, X_val_texts, y_train, y_val = train_test_split(
        texts, labels,
        test_size=0.2,
        random_state=42,
        stratify=labels  # Maintain class distribution
    )

    # Further split the validation set into validation (70%) and test (30%) sets
    X_val_texts, X_test_texts, y_val, y_test = train_test_split(
        X_val_texts, y_val,
        test_size=0.3,
        random_state=42,
        stratify=y_val  # Maintain class distribution
    )

    # Initialize TF-IDF vectorizer for character-level analysis
    vectorizer = TfidfVectorizer(
        analyzer='char',       # Character-level n-grams
        ngram_range=(1,5),     # Use n-grams from size 1 to 5
        max_features=500,      # Use top 500 features only
        lowercase=True         # Convert all text to lowercase
    )

    # Fit the vectorizer on the training texts and transform them to numeric features
    X_train = vectorizer.fit_transform(X_train_texts)

    # Transform validation and test texts using the same vectorizer
    X_val = vectorizer.transform(X_val_texts)
    X_test = vectorizer.transform(X_test_texts)

    # Return processed data and the fitted vectorizer
    return X_train, y_train, X_val, y_val, X_test, y_test, vectorizer